from .txt_utils import scrape_data, sort_scraped
from ...text_search import Classify


class TextApi:
    def __init__(self, config):
        data, save_timestamp = scrape_data(config)
        classifier = Classify()
        sort_scraped(config, classifier, data, save_timestamp)



from datetime import datetime
import requests
import time
import json


class RedditScraper:
    def __init__(self, config):
        if "author" not in config and "subreddit" not in config:
            raise ValueError("Fill in either username or subreddit")
        self.filter_string = "&"
        for k, v in config.items():
            if v:
                self.filter_string += str(k) + "=" + str(v) + "&"

    def scrape(self, object_type, start_time=None, end_time=None, limit=1000):
        data = []
        count = 0
        previous_epoch = int(datetime.utcnow().timestamp())

        limit_flag = False

        url = "https://api.pushshift.io/reddit/{}/search?limit={}&sort=desc{}before={}"
        if start_time:
            url = "https://api.pushshift.io/reddit/{}/search?limit={}&sort=desc{}before={}&after={}"
        while True:
            if start_time:
                if end_time:
                    new_url = url.format(
                        object_type, limit, self.filter_string, end_time, start_time
                    )
                else:
                    new_url = url.format(
                        object_type,
                        limit,
                        self.filter_string,
                        str(previous_epoch),
                        start_time,
                    )
            else:
                new_url = url.format(
                    object_type, limit, self.filter_string, str(previous_epoch)
                )
            json_text = requests.get(
                new_url, headers={
                    "User-Agent": "Post downloader by /u/Watchful1"}
            )
            time.sleep(
                1
            )  # pushshift has a rate limit, if we send requests too fast it will start returning error messages
            try:
                json_data = json_text.json()
            except json.decoder.JSONDecodeError:
                time.sleep(1)
                print("JSONDecodeError")
                continue
            if "data" not in json_data:
                break
            objects = json_data["data"]
            if len(objects) == 0:
                break
            for object in objects:
                previous_epoch = object["created_utc"] - 1
                if end_time:
                    end_time = previous_epoch
                count += 1
                if object_type == "comment":
                    try:
                        data.append(
                            object["body"]
                            .encode(encoding="ascii", errors="ignore")
                            .decode()
                        )
                        if len(data) >= limit:
                            print(
                                "Subreddit limit reached, stopped collecting comments..."
                            )
                            limit_flag = True
                    except Exception as err:

                        print("Couldn't get comment")
                        print(err)
                elif object_type == "submission":
                    if object["is_self"]:
                        if "selftext" not in object:
                            continue
                        try:
                            data.append(
                                object["selftext"]
                                .encode(encoding="ascii", errors="ignore")
                                .decode()
                            )
                            if len(data) >= limit:
                                print(
                                    "Subreddit limit reached, stopped collecting submissions..."
                                )
                                limit_flag = True
                        except Exception as err:
                            print("Couldn't get post")
                            print(err)

            print(
                "Saved {} {}s through {}".format(
                    len(data),
                    object_type,
                    datetime.fromtimestamp(previous_epoch).strftime("%Y-%m-%d"),
                )
            )

            if limit_flag:
                break
        return data



import os
import re
import requests
import praw
import configparser
import concurrent.futures
import argparse
from datetime import datetime, timedelta


class redditImageScraper:
    def __init__(self, sub, limit, order, client_id, client_secret, user_agent, nsfw=False, img_datastore="images"):
        config = configparser.ConfigParser()
        config.read('conf.ini')
        self.sub = sub
        self.limit = limit
        self.order = order
        self.nsfw = nsfw
        self.path = f'{img_datastore}/{self.sub}/'
        self.reddit = praw.Reddit(
            client_id=client_id, client_secret=client_secret, user_agent=user_agent)

    def download(self, image):
        r = requests.get(image['url'])
        with open(image['fname'], 'wb') as f:
            f.write(r.content)

    def start(self, start_time=None, remove_existing=True):
        images = []
        start_time = datetime.timestamp(start_time)
        go = 0
        if self.order == 'hot':
            submissions = self.reddit.subreddit(self.sub).hot(limit=None)
        elif self.order == 'top':
            submissions = self.reddit.subreddit(self.sub).top(limit=None)
        elif self.order == 'new':
            submissions = self.reddit.subreddit(self.sub).new(limit=None)

        for submission in submissions:
            if not submission.stickied and submission.over_18 == self.nsfw \
                    and submission.url.endswith(('jpg', 'jpeg', 'png')):
                if start_time:
                    created_time = datetime.utcfromtimestamp(
                        submission.created_utc)
                    created_time = datetime.timestamp(created_time)

                    if created_time >= start_time:
                        fname = self.path + \
                            re.search('(?s:.*)\w/(.*)', submission.url).group(1)
                        if not os.path.isfile(fname):
                            images.append(
                                {'url': submission.url, 'fname': fname})
                            go += 1
                            if go >= self.limit:
                                break
                else:
                    fname = self.path + \
                        re.search('(?s:.*)\w/(.*)', submission.url).group(1)
                    if not os.path.isfile(fname):
                        images.append({'url': submission.url, 'fname': fname})
                        go += 1
                        if go >= self.limit:
                            break
        if len(images):
            if remove_existing:
                if os.path.exists(self.path):
                    for file in os.listdir(self.path):
                        os.remove(os.path.join(self.path, file))
                    os.rmdir(self.path)
            if not os.path.exists(self.path):
                os.makedirs(self.path)
            with concurrent.futures.ThreadPoolExecutor() as ptolemy:
                ptolemy.map(self.download, images)



import pandas as pd
from datetime import datetime
from ...text_search import RedditScraper, SubredditSearch
import os


def scrape_data(config, write_to_file="json"):

    save_timestamp = str(int(datetime.now().timestamp()))
    if "subreddit_search_term" not in config.keys():
        if "subreddits" not in config.keys():
            raise AttributeError(
                "Add either 'subreddit_search_term' or 'subreddits' to config"
            )

    object_type = config["subreddit_object_type"]

    if config["start_time"]:
        start_time = datetime.strptime(config["start_time"],
                                       "%d.%m.%Y %H:%M:%S")
    else:
        start_time = None
    if config["end_time"]:
        end_time = datetime.strptime(config["end_time"], "%d.%m.%Y %H:%M:%S")
    else:
        end_time = None

    if "subreddits" in config.keys():
        subreddits = config["subreddits"]

    elif "subreddit_search_term" in config.keys():
        sub_search = SubredditSearch()
        subreddit_search_term = config["subreddit_search_term"]
        print("searching for subreddits......")
        subreddits = sub_search.search(
            subreddit_search_term, object_type,
            start_time=start_time, end_time=end_time
        )

    data = []
    subreddit_names = []
    finished_rows = 0

    resume_task_timestamp = None
    if "resume_task_timestamp" in config.keys():
        resume_task_timestamp = config["resume_task_timestamp"]
    # check for finished subreddit files

    if resume_task_timestamp:

        resume_task_timestamp = str(resume_task_timestamp)
        files = os.listdir(resume_task_timestamp)
        finished_subreddits = [i.replace(".json", "")
                               for i in files if i.endswith(".json")]
        finished_rows = sum([len(pd.read_json(
            os.path.join(resume_task_timestamp, i)
        )) for i in files if i.endswith(".json")]
        )
        config["total_limit"] = config["total_limit"] - finished_rows

        subreddits = [i for i in subreddits if i not in finished_subreddits]
        save_timestamp = resume_task_timestamp

        print("resuming from checkpoint "
              + resume_task_timestamp +
              "..... " + str(finished_rows) + "data points processed already")

    for subreddit in subreddits:
        print(subreddit)
        config["subreddit"] = subreddit
        r = RedditScraper(config)
        try:
            subreddit_data = r.scrape(
                object_type,
                start_time=start_time,
                end_time=end_time,
                limit=config["subreddit_text_limit"],
            )

            subreddit_names += [subreddit for i in range(len(subreddit_data))]
            data += subreddit_data

            if write_to_file:

                if not os.path.exists(save_timestamp):
                    os.mkdir(save_timestamp)

                pd.DataFrame({"text": data,
                              "subreddit": subreddit_names}).to_json(
                    os.path.join(save_timestamp, subreddit + ".json")
                )

            print(str(len(data)+finished_rows)+" data points collected")
            if len(data) >= config["total_limit"]:
                break
        except Exception as e:
            print(e)
            continue

    scraped_outputs = pd.DataFrame({"text": data, "subreddit": subreddit_names})

    if write_to_file:
        scraped_outputs.to_json(os.path.join(save_timestamp,
                                             "scraped_full.json"),
                                orient="records",
                                lines=True)

    return scraped_outputs, save_timestamp


def sort_scraped(config, classifier, data, save_timestamp):

    def _clean_data(data):

        remove_terms = ["", "[removed]"]
        out = data[~data['text'].isin(remove_terms)]

        return out

    if "subreddit_search_term" in config.keys():
        data = _clean_data(data)
        print(data, len(data))
        subreddit_search_term = [config["subreddit_search_term"]]
        out = classifier.return_tag(
            list(data["text"]), subreddit_search_term,
        )
        pd.DataFrame(out).to_json(os.path.join(save_timestamp,
                                               "scraped_classified.json"),
                                  orient="records",
                                  lines=True)



from ...image_search import redditImageScraper, ImageSearch
from datetime import datetime
import os
from ...text_search import SubredditSearch


def scrape_images(config, image_datastore_dir):

    if "subreddit_search_term" in config.keys() and "subreddits" in config.keys():
        raise AttributeError(
            "Add one of 'subreddit_search_term' or 'subreddits' to config")

    if "subreddit_search_term" not in config.keys() and "subreddits" not in config.keys():
        raise AttributeError(
            "Add either 'subreddit_search_term' or 'subreddits' to config")

    object_type = config["subreddit_object_type"]

    if 'start_time' in config.keys():
        start_time = datetime.strptime(config["start_time"],
                                       '%d.%m.%Y %H:%M:%S')
    else:
        start_time = None

    if 'end_time' in config.keys():
        end_time = datetime.strptime(config["end_time"],
                                     '%d.%m.%Y %H:%M:%S')
    else:
        end_time = None

    if "subreddits" in config.keys():
        subreddits = config["subreddits"]
    elif "subreddit_search_term" in config.keys():
        sub_search = SubredditSearch()
        subreddit_search_term = config["subreddit_search_term"]
        subreddits = sub_search.search(
            subreddit_search_term,
            object_type,
            start_time=start_time,
            end_time=end_time)
    # count total images which got scraped
    total_images = 0
    for subreddit in subreddits:
        img_scraper = redditImageScraper(subreddit,
                                         config["subreddit_image_limit"],
                                         config["sort_by"],
                                         client_id=config["client_id"],
                                         client_secret=config["client_secret"],
                                         user_agent=config["user_agent"],
                                         img_datastore=image_datastore_dir)
        if end_time:
            pass  # write code to stop cron job when date is end time
        img_scraper.start(start_time=start_time)
        total_images += config["subreddit_image_limit"]
        if total_images >= config["total_limit"]:
            subreddits = [i for i in os.listdir(
                image_datastore_dir) if i in subreddits]
            break
    return subreddits, image_datastore_dir


def image_classify(config, subreddits, img_datastore):
    print('subreddits {}'.format(subreddits))
    output_paths = []

    for subreddit in subreddits:
        search = ImageSearch(img_dir=os.path.join(
            os.getcwd(), img_datastore+"/"+subreddit))
        output_paths += search.search(
            config["subreddit_search_term"],
            k=config["subreddit_image_limit"],
            threshold_score=0.2)

    return output_paths



class Classify:

    def __init__(self, classifier_algorithm="zero_shot", model=None):
        self.classifier_algorithm = classifier_algorithm

        if classifier_algorithm == "cosine_similarity":
            if not model:
                model = "paraphrase-multilingual-MiniLM-L12-v2"
            from sentence_transformers import SentenceTransformer
            self.model = SentenceTransformer(model)
        else:
            from transformers import pipeline
            if not model:
                model = "cross-encoder/nli-deberta-v3-small"
            self.model = pipeline(task="zero-shot-classification", model=model)

    def return_tag(self, textlist, labels, threshold_score=0.5):

        if self.classifier_algorithm == "cosine_similarity":
            from sentence_transformers import util
            text_embeds = self.model.encode(textlist, convert_to_tensor=True)
            label_embeds = self.model.encode(labels, convert_to_tensor=True)
            cosine_scores = util.pytorch_cos_sim(text_embeds, label_embeds)
            text = []
            classes = []

            for index in range(len(cosine_scores)):
                for i in range(len(cosine_scores[index])):

                    if cosine_scores[index][i] > threshold_score:
                        text.append(textlist[index])
                        classes.append(labels[i])

            return {"text": text, "label": classes}

        else:

            pipe = self.model
            labels += ["other"]
            res = pipe(textlist, labels)
            text = []
            classes = []

            for i in range(len(res)):

                if max(res[i]["scores"]) > threshold_score:
                    label = res[i]["labels"]
                    [res[i]["scores"].index(max(res[i]["scores"]))]

                    if label != "other":
                        text.append(res[i]["sequence"])
                        classes.append(label)

            return {"text": text, "label": classes}



from .img_utils import scrape_images, image_classify
import os
import shutil


class ImageApi:
    def __init__(self, config, image_datastore_dir='image_datastore',
                 dest_dir='filtered_images'):
        subreddits, img_datastore = scrape_images(config, image_datastore_dir)
        output_paths = image_classify(config, subreddits, img_datastore)

        if os.path.exists(dest_dir):
            shutil.rmtree(dest_dir)

        os.mkdir(dest_dir)

        for path in output_paths:
            shutil.copy(path, dest_dir)



import requests


class SubredditSearch:
    def search(self, query, object_type, start_time=None, end_time=None):
        if start_time:
            if end_time:
                url = "https://api.pushshift.io/reddit/search/{}/?q={}&before={}&after={}"
                url = url.format(object_type, query, end_time, start_time)
            else:
                url = "https://api.pushshift.io/reddit/search/{}/?q={}&after={}"
                url = url.format(object_type, query, start_time)
        else:
            url = "https://api.pushshift.io/reddit/search/{}/?q={}"
            url = url.format(object_type, query)
        json_text = requests.get(
            url, headers={'User-Agent': "Abhijith"}, verify=False).json()
        subreddits = [data["subreddit"] for data in json_text["data"]]
        return subreddits



from sentence_transformers import SentenceTransformer, util
import pickle
from PIL import Image
import os


class ImageSearch:
    def __init__(self, model_name='clip-ViT-B-32', img_dir="images"):
        self.model = SentenceTransformer(model_name)
        self.img_dir = img_dir

    def convert_images(self):
        for filename in os.listdir(self.img_dir):
            if filename.endswith((".png")):
                im1 = Image.open(os.path.join(self.img_dir, filename))
                im1 = im1.convert('RGB')
                im1.save(os.path.join(self.img_dir,
                         filename.replace(".png", ".jpg")))
                os.remove(os.path.join(self.img_dir, filename))

    def load_image_paths(self, folder):
        paths = []
        for filename in os.listdir(folder):
            if filename.endswith((".jpg", "png", "jpeg")):
                paths.append(os.path.join(folder, filename))
        return paths

    def encode_images(self, dirname):
        self.convert_images()
        img_names = self.load_image_paths(dirname)
        img_emb = self.model.encode([Image.open(filepath)
                                     for filepath in img_names],
                                    batch_size=128,
                                    convert_to_tensor=True,
                                    show_progress_bar=True)
        return img_emb, img_names

    def search(self, query, k=3, use_precomputed_embeddings=False,
               threshold_score=0.2):
        if use_precomputed_embeddings:
            emb_filename = 'unsplash-25k-photos-embeddings.pkl'
            # Download dataset if does not exist
            if not os.path.exists(emb_filename):
                util.http_get('http://sbert.net/datasets/' +
                              emb_filename, emb_filename)

            with open(emb_filename, 'rb') as fIn:
                img_names, img_emb = pickle.load(fIn)
            print("Images:", len(img_names))
        else:
            img_emb, img_names = self.encode_images(self.img_dir)
        query_emb = self.model.encode(
            [query], convert_to_tensor=True, show_progress_bar=False)
        hits = util.semantic_search(query_emb, img_emb, top_k=k)[0]
        matched_images = []
        for hit in hits:
            if hit["score"] > threshold_score:
                matched_images.append(img_names[hit['corpus_id']])
        return matched_images



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = ["sentence_transformers", "praw"]
setuptools.setup(
    name="redditflow",
    version="0.0.1",
    author="Abhijith Neil Abraham",
    author_email="abhijithneilabrahampk@gmail.com",
    description="Data Curation over Time",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT License',
    url="https://github.com/nfflow/reflow",
    install_requires=install_requires,
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    include_package_data=True
)
